/*
** tlsfifo.c for tlsinject in
**
** Made by Averous Julien-Pierre
** E-Mail   <j.averous@sourcemac.com>
**
** Started on  Fri Dec  5 14:43:18 2008 Averous Julien-Pierre
** Last update Sat Dec  6 19:44:20 2008 Averous Julien-Pierre
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "tlsfifo.h"

/*!
** Init a fifo object
**
** @param fifo The fifo queue
** @param freer The function to free content
*/
void tls_fifo_init(tls_fifo *fifo, fifo_free_item freer)
{
  if (!fifo)
    return;

  pthread_mutex_init(&(fifo->mutex), NULL);

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  fifo->first = NULL;
  fifo->last = NULL;
  fifo->total_size = 0;

  fifo->free_foo = freer;

  // Unlock]
  pthread_mutex_unlock(&(fifo->mutex));
}

/*!
** Free the fifo queue (use the user function to free content)
**
** @param fifo The fifo queue
*/
void tls_fifo_clean(tls_fifo *fifo)
{
  if (!fifo)
    return;

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  tls_fifo_item *item = fifo->first;
  tls_fifo_item *next;

  while (item)
  {
    if (fifo->free_foo)
      fifo->free_foo(fifo, item->data, item->data_len);
    next = item->next;

    free(item);
    item = next;
  }

  fifo->first = NULL;
  fifo->last = NULL;
  fifo->total_size = 0;

  // Unlock by destroy]
  pthread_mutex_destroy(&(fifo->mutex));
}

/*!
** Push data in the fifo queue
**
** @param fifo The fifo queue
** @param data The data to push
** @param size The size of the data to push (used for tls_fifo_size)
*/
void tls_fifo_push(tls_fifo *fifo, void *data, unsigned size)
{
  if (!fifo)
    return;

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  tls_fifo_item *item = malloc(sizeof(tls_fifo_item));

  item->data = data;
  item->data_len = size;

  if (!fifo->first || !fifo->last)
  {
    item->next = NULL;
    item->prev = NULL;
    fifo->first = item;
    fifo->last = item;
  }
  else
  {
    item->prev = fifo->last;
    item->next = NULL;
    fifo->last->next = item;
    fifo->last = item;
  }

  fifo->total_size += size;

  // Unlock]
  pthread_mutex_unlock(&(fifo->mutex));
}

/*!
** Pop data from the fifo queue
**
** @param fifo The fifo queue
** @param data The poped data (ignored if NULL)
** @param size The size of the poped data (ignored if NULL)
**
** @return 1 if data can be poped, 0 if the queue is empty
*/
int tls_fifo_pop(tls_fifo *fifo, void **data, unsigned *size)
{
  if (!fifo)
    return 0;

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  if (!fifo->first || !fifo->last)
  {
    // Unlock]
    pthread_mutex_unlock(&(fifo->mutex));
    return 0;
  }

  tls_fifo_item *item = fifo->first;

  if (item == fifo->last)
    fifo->last = NULL;

  if (item)
    *data = item->data;
  if (size)
    *size = item->data_len;

  fifo->total_size -= item->data_len;

  fifo->first = fifo->first->next;

  if (item->next)
    item->next->prev = NULL;

  free(item);

  // Unlock]
  pthread_mutex_unlock(&(fifo->mutex));

  return 1;
}

/*!
** Get the fifo content size
**
** @param fifo The fifo queue
**
** @return The user content size in the queue
*/
unsigned tls_fifo_size(tls_fifo *fifo)
{
  unsigned result = 0;

  if (!fifo)
    return 0;

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  result = fifo->total_size;

  // Unlock]
  pthread_mutex_unlock(&(fifo->mutex));

  return result;
}

/*!
** Print the fifo queue (print address for user data)
**
** @param fifo The fifo queue
*/
void tls_fifo_print(tls_fifo *fifo)
{
  if (!fifo)
    return;

  // [Lock
  pthread_mutex_lock(&(fifo->mutex));

  tls_fifo_item *item = fifo->first;

  printf("[");
  while (item)
  {
    unsigned dtptr = (unsigned)item->data;

    printf("(0x%08x, %u)", dtptr, item->data_len);
    item = item->next;
  }
  printf("]\n");

  // Unlock]
  pthread_mutex_unlock(&(fifo->mutex));
}
