/*
** tlshash.c for tlsinject in
**
** Made by Averous Julien-Pierre
** E-Mail   <j.averous@sourcemac.com>
**
** Started on  Wed Nov 26 13:55:54 2008 Averous Julien-Pierre
** Last update Sat Dec  6 19:40:18 2008 Averous Julien-Pierre
*/

#include <stdlib.h>
#include <stdio.h>

#include "tlshash.h"

// Define the hash function for the hash table
#define HASH(Key) (((unsigned long long)(Key)) % TABLE_SIZE)

/*
** Internals Functions Prototypes
*/
static void _free_item(tls_hash *table, tls_hash_item *item);
static void _free_content(tls_hash *table, void *content);

/*!
** Init a hash table
**
** @param table : the hash table to init
*/
void tls_hash_init(tls_hash *table, hash_free_item freer)
{
  if (!table)
    return;

  // Use a mutex to protect our hash functions
  pthread_mutex_init(&(table->mutex), NULL);

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  // Init the table
  table->array = calloc(TABLE_SIZE, sizeof(tls_hash_item *));

  // Remain the free content function
  table->free_foo = freer;

  // Unlock]
  pthread_mutex_unlock(&(table->mutex));
}


/*!
** Clean a hash table (free items via the user function)
**
** @param table : the hash table to clean
*/
void tls_hash_clean(tls_hash *table)
{
  if (!table)
    return;

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  // Free the table
  if (table->array)
  {
    unsigned i;

    for (i = 0; i < TABLE_SIZE; i++)
    {
      tls_hash_item *item = table->array[i];
      tls_hash_item *next = NULL;

      while (item)
      {
	next = item->next;
	_free_item(table, item);
	item = next;
      }
    }
    free(table->array);
    table->array = NULL;
  }

  // Unlock by destroy]
  pthread_mutex_destroy(&(table->mutex));
}

/*!
** Associate an object to a key on the hash table
** If the object already exist, it is replaced by the new object (the old object is freed by the user function)
**
** @param table The hash table
** @param key The key of the object to set
** @param content The object to set
*/
void tls_hash_set_content(tls_hash *table, long long key, void *content)
{
  if (!table || !table->array)
    return;

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  unsigned	index = HASH(key);

  // Set the hash item in the hash table
  if (table->array[index])
  {
    tls_hash_item *run = table->array[index];

    while (run)
    {
      // If the key already exist, change the content
      if (run->key == key)
      {
	_free_content(table, run->content);
	run->content = content;
	break;
      }

      // If we are in end-of-chain, append the new item
      if (run->next == NULL)
      {
	tls_hash_item *item = malloc(sizeof(tls_hash_item));

	// Init the new hash item
	item->key = key;
	item->content = content;
	item->next = NULL;

	// Set the end of the chain to this new item
	run->next = item;
	break;
      }
      run = run->next;
    }
  }
  else
  {
    tls_hash_item *item = malloc(sizeof(tls_hash_item));

    // Init the new hash item
    item->key = key;
    item->content = content;
    item->next = NULL;

    // Set the object in the table
    table->array[index] = item;
  }

  // Unlock]
  pthread_mutex_unlock(&(table->mutex));
}

/*!
** Retrieve an object with a key
**
** @param table The hash table
** @param key The key of the object to retrieve
**
** @return The object if the key exist, NULL else
*/
void *tls_hash_get_content(tls_hash *table, long long key)
{
  if (!table || !table->array)
    return NULL;

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  unsigned	index = HASH(key);

  // Search the key
  if (table->array[index])
  {
    tls_hash_item *run = table->array[index];

    while (run)
    {
      if (run->key == key)
      {
	pthread_mutex_unlock(&(table->mutex));
	return run->content;
      }
      run = run->next;
    }
  }

  // Unlock]
  pthread_mutex_unlock(&(table->mutex));
  return NULL;
}

/*!
** Remove an object with its key (the object is freed with the user function)
**
** @param table The hash table
** @param key The key of the object to remove
*/
void tls_hash_remove(tls_hash *table, long long key)
{
  if (!table || !table->array)
    return;

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  unsigned	index = HASH(key);

  if (table->array[index])
  {
    tls_hash_item *item = table->array[index];

    // If the remove key is on the array
    if (item->key == key)
    {
      table->array[index] = item->next;

      _free_item(table, item);
    }
    // else search on the linked list
    else
    {
      while (item)
      {
	if (item->next && item->next->key == key)
	{
	  _free_item(table, item->next);
	  item->next = item->next->next;
	  break;
	}
	item = item->next;
      }
    }
  }

  // Unlock]
  pthread_mutex_unlock(&(table->mutex));
}

/*!
** Print the content of the hash table (print address for user object)
**
** @param table The hash table
*/
void tls_hash_print(tls_hash *table)
{
  if (!table || !table->array)
    return;

  unsigned i;

  // [Lock
  pthread_mutex_lock(&(table->mutex));

  printf("|------\n");
  for (i = 0; i < TABLE_SIZE; i++)
  {
    if (table->array[i])
    {
      tls_hash_item *item = table->array[i];

      while (item)
      {
	printf("[Key=%qd; Content=%p]->", item->key, item->content);
	item = item->next;
      }
      printf("X\n");
    }
  }
  printf("|------\n");

  // Unlock]
  pthread_mutex_unlock(&(table->mutex));
}


/*
** Internals Functions Implementations
*/
/*!
** Free an internal item on the hash_table
**
** @param table The hash table
** @param item The internal item to free
*/
static void _free_item(tls_hash *table, tls_hash_item *item)
{
  if (item)
    _free_content(table, item->content);

  free(item);
}

/*!
** Free an user object (call the user function)
**
** @param table The hash table
** @param content The user object to free
*/
static void _free_content(tls_hash *table, void *content)
{
  if (table->free_foo)
    table->free_foo(table, content);
}
