/*
** tlsinject.c for tlsinject in
**
** Made by Averous Julien-Pierre
** E-Mail   <j.averous@sourcemac.com>
**
** Started on  Wed Nov 26 13:54:40 2008 Averous Julien-Pierre
** Last update Mon Dec  8 04:01:24 2008 Averous Julien-Pierre
*/

/*
** Includes
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>

#include <openssl/ssl.h>

#include "tlshash.h"
#include "tlsfifo.h"
#include "tlscommon.h"

/*
** Global Variables
*/
tls_hash hash_table;

/*
** GLobal Define
*/
#define KEYFILE_CLIENT "/openssl_files/client.pem"
#define PASSWORD_CLIENT "password"

#define KEYFILE_SERVER "/openssl_files/server.pem"
#define PASSWORD_SERVER "password"
#define DHFILE_SERVER "/openssl_files/dh1024.pem"

/*
** Global Macro Functions
*/
#define MIN(A,B) (((A)<(B)) ? (A) : (B))

/*
** Debug Function
*/
#define DEBUG 1

#if DEBUG
# define DPRINT(Str) fputs(Str, stderr)
# define DFPRINT(Str, ...) fprintf(stderr, Str, __VA_ARGS__)
#else
# define DPRINT(Str)
# define DFPRINT(Str, ...)
#endif

/*
** Interlal Functions Prototypes
*/
static void free_associate(tls_hash *table, void *content);
static void free_content_dtls(tls_hash *table, void *content);
static void free_content_fifo(tls_fifo *fifo, void *data, unsigned size);

static int dgram_write(BIO *b, const char *in, int inl);
static int dgram_read(BIO *b, char *out, int outl);
static int dgram_puts(BIO *bp, const char *str);

static int sock_write(BIO *b, const char *in, int inl);
static int sock_read(BIO *b, char *out, int outl);
static int sock_puts(BIO *bp, const char *str);

static inline long long (peer_key)(struct sockaddr_in *address);

/*
** Structures for the TLS/DTLS connection
*/
/// Enumeration of connection type (obtained via the socket() function)
typedef enum
{
  ssl_type_tcp,			/*!< The connection is on TCP*/
  ssl_type_udp			/*!< The connection is on UDP*/
} ssl_type;

/// Contain information about a TLS tunnel
typedef struct tls_content_
{
  SSL_CTX	*ctx;		/*!< The SSL context*/
  SSL		*ssl;		/*!< The SSL session object*/
  BIO		*sbio;		/*!< The SSL I/O object*/
  unsigned char	session_id;	/*!< The session ID of the TLS tunnel*/
} tls_content;

/// Contain information about a DTLS tunnel
typedef struct dtls_content_
{
  SSL_CTX	*ctx;		/*!< The SSL context*/
  SSL		*ssl;		/*!< The SSL session object*/
  BIO		*sbio;		/*!< The SSL I/O object*/
  unsigned char	session_id;	/*!< The session ID of the DTLS tunnel*/

  tls_fifo	fifo;		/*!< A FIFO queue, like a buffer, of trame of the connexion*/

  char		remote[1024];  	/*!< The remote peer address object, obtained by recvfrom*/
  socklen_t	remote_len;	/*!< The address object size*/
} dtls_content;

static dtls_content *dtls_read(int socket);

/// Allow to maintain a link between a TLS/DTLS connection and a user socket
typedef struct ssl_associate_
{
  // Content type
  ssl_type	type;		/*!< The type of the socket (UDP/TCP)*/
  int skip;

  // Content
  tls_content	*tls;		/*!< Only in TCP type, contain the TLS tunnel information*/
  tls_hash	*dtls;		/*!< Only in UDP type, contain an "array" of DTLS tunnel (multiplexed socket), indentified by each IPv4+Port of each peer*/

  // Address
  char		addr[1024];
  socklen_t	addr_len;
} ssl_associate;


/// The internal BIO data object of OpenSSL 0.9.8, taken from the bss_dgram.c file. Be aware of change in next version of OpenSSL.
typedef struct bio_dgram_data_st
{
  struct sockaddr peer;
  unsigned int connected;
  unsigned int _errno;
  unsigned int mtu;
} bio_dgram_data;


/*!
** The library contructor : called on library load (dlopen, etc.)
**
*/
void __attribute__((constructor)) init_lib()
{
  // Banner
  fprintf(stderr, "> Welcome to the tls inject library, a module for the Asterisk SRS-EPITA project <\n");
  fprintf(stderr, "> Wrote by Averous Julien-Pierre <\n\n");

  // Init the hash table (used to associate a socket with a tls session)
  tls_hash_init(&hash_table, free_associate);

  BIO_METHOD *method;

  // Change the bio-socket function to use our bridge to the libc
  method = BIO_s_socket();
  method->bwrite = sock_write;
  method->bread = sock_read;
  method->bputs = sock_puts;

  // Change the bio-dgram-socket to use our bridge to the libc, and to manage multiplex socket
  method = BIO_s_datagram();
  method->bwrite = dgram_write;
  method->bread = dgram_read;
  method->bputs = dgram_puts;
}

/*
** Bridge to libc
** See the man for information
*/
int my_socket(int domain, int type, int protocol)
{
  static int (*lib_socket)(int, int, int) = NULL;

  if (!lib_socket)
    lib_socket = (int (*)(int, int, int))dlsym(RTLD_NEXT, "socket");

  return lib_socket(domain, type, protocol);
}

int my_setsockopt(int socket, int level, int option_name, const void *option_value, socklen_t option_len)
{
  static int (*lib_setsockopt)(int, int, int, const void *, socklen_t) = NULL;

  if (!lib_setsockopt)
    lib_setsockopt = (int (*)(int, int, int, const void *, socklen_t))dlsym(RTLD_NEXT, "setsockopt");

  return lib_setsockopt(socket, level, option_name, option_value, option_len);
}

int my_bind(int socket, const struct sockaddr *address, socklen_t address_len)
{
  static int (*lib_bind)(int, const struct sockaddr *, socklen_t) = NULL;

  if (!lib_bind)
    lib_bind = (int (*)(int, const struct sockaddr *, socklen_t))dlsym(RTLD_NEXT, "bind");

  return lib_bind(socket, address, address_len);
}

int my_listen(int socket, int backlog)
{
  static int (*my_listen)(int, int) = NULL;

  if (!my_listen)
    my_listen = (int (*)(int, int))dlsym(RTLD_NEXT, "listen");

  return my_listen(socket, backlog);
}

int my_connect(int socket, const struct sockaddr *address, socklen_t address_len)
{
  static int (*lib_connect)(int, const struct sockaddr *, socklen_t) = NULL;

  if (!lib_connect)
    lib_connect = (int (*)(int, const struct sockaddr *, socklen_t))dlsym(RTLD_NEXT, "connect");

  return lib_connect(socket, address, address_len);
}

int my_accept(int socket, struct sockaddr *address, socklen_t *address_len)
{
  static int (*lib_accept)(int, struct sockaddr *, socklen_t *) = NULL;

  if (!lib_accept)
    lib_accept = dlsym(RTLD_NEXT, "accept");

  return lib_accept(socket, address, address_len);
}

int my_close(int filedes)
{
  static int (*lib_close)(int) = NULL;

  if (!lib_close)
    lib_close = dlsym(RTLD_NEXT, "close");

  return lib_close(filedes);
}

ssize_t my_recvmsg(int socket, struct msghdr *message, int flags)
{
  static ssize_t (*lib_recvmsg)(int, struct msghdr *, int) = NULL;

  if (!lib_recvmsg)
    lib_recvmsg = dlsym(RTLD_NEXT, "recvmsg");

  return lib_recvmsg(socket, message, flags);
}

ssize_t my_recvfrom(int socket, void *buffer, size_t length, int flags, struct sockaddr *address, socklen_t *address_len)
{
  static ssize_t (*lib_recvfrom)(int, void *, size_t, int, struct sockaddr *, socklen_t *) = NULL;

  if (!lib_recvfrom)
    lib_recvfrom = dlsym(RTLD_NEXT, "recvfrom");

  return lib_recvfrom(socket, buffer, length, flags, address, address_len);
}

ssize_t my_sendmsg(int socket, const struct msghdr *buffer, int flags)
{
  static ssize_t (*lib_sendmsg)(int, const struct msghdr *, int) = NULL;

  if (!lib_sendmsg)
    lib_sendmsg = dlsym(RTLD_NEXT, "sendmsg");

  return lib_sendmsg(socket, buffer, flags);
}

ssize_t my_sendto(int socket, const void *buffer, size_t length, int flags, const struct sockaddr *dest_addr, socklen_t dest_len)
{
  static ssize_t (*lib_sendto)(int, const void *, size_t, int, const struct sockaddr *, socklen_t) = NULL;

  if (!lib_sendto)
    lib_sendto = dlsym(RTLD_NEXT, "sendto");

  return lib_sendto(socket, buffer, length, flags, dest_addr, dest_len);
}


ssize_t my_read(int fildes, void *buf, size_t nbyte)
{
  static ssize_t(*lib_read)(int, void *, size_t) = NULL;

  if (!lib_read)
    lib_read = dlsym(RTLD_NEXT, "read");

  return lib_read(fildes, buf, nbyte);
}

ssize_t my_write(int fildes, const void *buf, size_t nbyte)
{
  static ssize_t(*lib_write)(int, const void *, size_t) = NULL;

  if (!lib_write)
    lib_write = dlsym(RTLD_NEXT, "write");

  return lib_write(fildes, buf, nbyte);
}

/*
** Inject
*/
int socket(int domain, int type, int protocol)
{
  int result =  my_socket(domain, type, protocol);

  DFPRINT("\n[[[socket (%i, %i, %i) => %i]]]\n", domain, type, protocol, result);

  // The injection take place only on ARPA Internet Protocols ...
  if (domain == AF_INET || domain == AF_UNIX)
  {
    // ... and for SOCK_STREAM ...
    if (type == SOCK_STREAM)
    {
      DFPRINT("*** tcp socket (%i)\n", result);
      ssl_associate *aso = calloc(1, sizeof(ssl_associate));

      aso->type = ssl_type_tcp;
      aso->tls = calloc(1, sizeof(tls_content));

      tls_hash_set_content(&hash_table, result, aso);
    }

    // ... or SOCK_DGRAM.
    else if (type == SOCK_DGRAM)
    {
      DFPRINT("*** udp socket (%i)\n", result);
      ssl_associate *aso = calloc(1, sizeof(ssl_associate));

      aso->type = ssl_type_udp;
      aso->dtls = malloc(sizeof(tls_hash));
      tls_hash_init(aso->dtls, free_content_dtls);

      tls_hash_set_content(&hash_table, result, aso);
    }
  }

  return result;
}

int setsockopt(int socket, int level, int option_name, const void *option_value, socklen_t option_len)
{
  DFPRINT("\n[[[setsockopt (%i, %i, %i)]]]\n", socket, level, option_name);

  return my_setsockopt(socket, level, option_name, option_value, option_len);
}

int bind(int socket, const struct sockaddr *address, socklen_t address_len)
{
  struct sockaddr_in *addr = (struct sockaddr_in *)address;

  DFPRINT("\n[[[bind (%i)]]]\n", socket);

  if (address_len == sizeof(struct sockaddr_in))
    DFPRINT("-> bind to %s:%u\n", inet_ntoa(addr->sin_addr), ntohs(addr->sin_port));

  ssl_associate *aso = tls_hash_get_content(&hash_table, socket);

  if (aso)
  {
    memcpy((void*)&aso->addr, (void*)address, address_len);
    aso->addr_len = address_len;
  }

  return my_bind(socket, address, address_len);
}

int listen(int socket, int backlog)
{
  DFPRINT("\n[[[listen (%i)]]]\n", socket);

  return my_listen(socket, backlog);
}

int connect(int socket, const struct sockaddr *address, socklen_t address_len)
{
  DFPRINT("\n[[[connect (%i)]]]\n", socket);

  ssl_associate *aso = tls_hash_get_content(&hash_table, socket);

  // If this socket was not build by us, do a libc connect
  if (!aso)
    return my_connect(socket, address, address_len);

  // Copy the connect address
  memcpy((void*)&aso->addr, (void*)address, address_len);
  aso->addr_len = address_len;

  if (aso->type == ssl_type_tcp)
  {
    tls_content	*tls = aso->tls;

    // If we don't have a tls context, there is a (strange) error
    if (!tls)
      return -1;

    // Do libc socket connection
    int result = my_connect(socket, address, address_len);
    int err = errno;

    // Check error and manage the case of non-blocking socket
    if (result == -1 && !(err == EWOULDBLOCK || err == EINPROGRESS))
      return result;

    // Build an SSL context
    tls->ctx = initialize_ctx(KEYFILE_CLIENT, PASSWORD_CLIENT, SSLv23_method());
    SSL_CTX_set_session_id_context(tls->ctx, &(tls->session_id), sizeof(tls->session_id));
    SSL_CTX_set_mode(tls->ctx, SSL_MODE_AUTO_RETRY);

    // Build the SSL tunnel
    tls->ssl = SSL_new(tls->ctx);

    // Build a BIO based on the user socket, and use it for read/write
    tls->sbio = BIO_new_socket(socket, BIO_NOCLOSE);
    SSL_set_bio(tls->ssl, tls->sbio, tls->sbio);

    // Do the SSL connexion
    if (SSL_connect(tls->ssl) <= 0 && !(err == EWOULDBLOCK || err == EINPROGRESS))
    {
      tls_hash_remove(&hash_table, socket);
      return -1;
    }

    return result;
  }
  else if (aso->type == ssl_type_udp)
  {
    return my_connect(socket, address, address_len);
  }

  return -1;
}

int accept(int socket, struct sockaddr *address, socklen_t *address_len)
{
  DFPRINT("\n[[[accept (%i)]]]\n", socket);

  // Until we are not a system error, try to accept a new connection (prevent SSL_accept error)
  while (1)
  {
    // Do libc socket accept
    int result = my_accept(socket, address, address_len);
    int err;

    // Check that the accept give us a new socket
    if (result == -1)
    {
      DPRINT("-> system accept error\n");
      return -1;
    }

    // Build a new associate entry for this new socket
    ssl_associate *aso = calloc(1, sizeof(ssl_associate));
    tls_content	  *tls = aso->tls;
    tls_hash_set_content(&hash_table, result, aso);

    // Set the type of tunnel
    aso->type = ssl_type_tcp;

    // Initialise the SSLcontext
    tls->ctx = initialize_ctx(KEYFILE_SERVER, PASSWORD_SERVER, SSLv23_method());
    SSL_CTX_set_mode(tls->ctx, SSL_MODE_AUTO_RETRY);

    // Set a Diffie Helman parameter
    load_dh_params(tls->ctx, DHFILE_SERVER);

    // Build the SSL tunnel
    tls->ssl = SSL_new(tls->ctx);

    // Build the BIO based on the new socket
    tls->sbio = BIO_new_socket(result, BIO_NOCLOSE);
    SSL_set_bio(tls->ssl, tls->sbio, tls->sbio);

    // Wait for SSL handshaking
    if ((err = SSL_accept(tls->ssl)) <= 0)
    {
      DPRINT("*** SSL Accept Handshake Failed\n");

      // Free the new entry
      tls_hash_remove(&hash_table, result);

      // If the error happen in the SSL handshake, wait to accept a new connection
      continue;
    }

    // All is right for the new socket : return it to the caller
    return result;
  }

  return -1;
}

int close(int filedes)
{
  DFPRINT("\n[[[close (%i)]]]\n", filedes);

  // Free the entry of this descriptor (if filedes is not on the hash table, we do nothing : by example if close is for a file and not a socket)
  // tls_hash will call free_associate for this object
  //tls_hash_remove(&hash_table, filedes);

  return my_close(filedes);
}

ssize_t recvmsg(int socket, struct msghdr *message, int flags)
{
  DFPRINT("\n[[[recvmsg (%i)]]]\n", socket);

  return my_recvmsg(socket, message, flags);
}

ssize_t recv(int socket, void *buffer, size_t length, int flags)
{
  return recvfrom(socket, buffer, length, flags, NULL, NULL);
}

ssize_t recvfrom(int socket, void *buffer, size_t length, int flags, struct sockaddr *address, socklen_t *address_len)
{
  DFPRINT("\n[[[recvfrom (%i)]]]\n", socket);

  ssl_associate *aso = tls_hash_get_content(&hash_table, socket);

  // If this not a "ssl socket", call libc function
  if (!aso || aso->skip)
    return my_recvfrom(socket, buffer, length, flags, address, address_len);

  // Check the socket type
  if (aso->type == ssl_type_tcp)
  {
    tls_content *tls = aso->tls;

    // If we don't have a ssl object, we have an error
    if (!tls || !tls->ssl)
      return -1;

    // Read data from the SSL session
    int r = SSL_read(tls->ssl, buffer, length);

    // Check the action result
    switch (SSL_get_error(tls->ssl, r))
    {
      // No error, return the size read
      case SSL_ERROR_NONE:
	return r;

	// End of data
      case SSL_ERROR_ZERO_RETURN:
	close(socket);
	return 0;
    }
  }
  else if (aso->type == ssl_type_udp)
  {
    // Retry recvfrom until we don't have a system error (prevent DTLS error)
    while (1)
    {
      // Get a dtls object (we should read a trame for this, because we use the IP and the Port to identify the peer : multiplexing of dtls session).
      // - If the dtls object doesn't exist (new peer), a new dtls_content object is build and it is added on the hash table of the aso.
      // - The trame received is pushed on the fifo of the dtls_content object to be used after.
      // - The address object of the remote peer is set on the dtls_content object.
      dtls_content *cdtls = dtls_read(socket);

      // If we don"t have a dtls_content (socket error, internal error on the aso, etc.) : exit
      DPRINT("-> check dtls\n");
      if (!cdtls)
	return -1;

      // "return" the address object to the user
      if (address)
	memcpy(address, cdtls->remote, cdtls->remote_len);
      if (address_len)
	*address_len = cdtls->remote_len;

      // Read data to the SSL session
      DPRINT("-> call SSL_read\n");
      int r = SSL_read(cdtls->ssl, buffer, length);

      // Check the SSL_read result
      DPRINT("-> call SSL_get_error\n");
      switch (SSL_get_error(cdtls->ssl, r))
      {
	// No error, return the size
	case SSL_ERROR_NONE:
	  return r;
      }

      DPRINT("-> error, reenter in receive stat\n");
    }
  }

  return -1;
}

ssize_t sendmsg(int socket, const struct msghdr *buffer, int flags)
{
  return my_sendmsg(socket, buffer, flags);
}

ssize_t send(int socket, const void *buffer, size_t length, int flags)
{
  return sendto(socket, buffer, length, flags, NULL, 0);
}

ssize_t sendto(int socket, const void *buffer, size_t length, int flags, const struct sockaddr *dest_addr, socklen_t dest_len)
{
  DFPRINT("\n[[[sendto (%u)]]]\n", socket);

  ssl_associate *aso = tls_hash_get_content(&hash_table, socket);
  struct sockaddr_in *addr = (struct sockaddr_in*)dest_addr;

  if (dest_len == sizeof(struct sockaddr_in) && ntohs(addr->sin_port) == 53)
  {
    DPRINT("-> DNS Port, use standard send-to\n");
    if (aso)
      aso->skip = 1;
  }

  // If this not a "ssl socket", call the libc function
  if (!aso || aso->skip)
    return my_sendto(socket, buffer, length, flags, dest_addr, dest_len);

  // Check the socket type
  if (aso->type == ssl_type_tcp)
  {
    tls_content *tls = aso->tls;

    // If we doesn't have a ssl object, we have an error
    if (!tls || !tls->ssl)
      return -1;

    // Write data to the SSL session
    int r = SSL_write(tls->ssl, buffer, length);

    // Check the SSL_write result
    switch (SSL_get_error(tls->ssl, r))
    {
      // No error, return the size
      case SSL_ERROR_NONE:
	return r;

	// End of data
      case SSL_ERROR_ZERO_RETURN:
	close(socket);
	return 0;
    }
  }
  else if (aso->type == ssl_type_udp)
  {
    tls_hash *dtls = aso->dtls;

    // Check that we have a dtls_content hash table
    DPRINT("-> check that we have a dtls in the aso\n");
    if (!dtls)
      return -1;

    // Check that the user give us a remote address object (mandatory to identify the remote peer, and so the dtls session)
    DPRINT("-> check that we have a target address\n");
    if (!dest_addr)
      return -1;

    // Check that this is a standard IPv4 address (the peer is indentified with a IPv4 and port).
    DPRINT("-> check the type of address structure\n");
    if (dest_len != sizeof(struct sockaddr_in) || dest_addr->sa_family != AF_INET)
      return -1;

    // Get the dtls_content of this peer (containing the dtls session information)
    DPRINT("-> get the cdtls of the target peer\n");
    struct sockaddr_in	*remote = (struct sockaddr_in *)dest_addr;
    long long  		id = peer_key(remote);//(int)remote->sin_addr.s_addr;
    dtls_content	*cdtls = tls_hash_get_content(dtls, id);

    // If the peer doesn't exist (first send), build a peer and initialize the dtls session
    if (!cdtls)
    {
      DPRINT("*** build a new peer\n");
      cdtls = calloc(1, sizeof(dtls_content));

      // Initialize the FIFO
      tls_fifo_init(&cdtls->fifo, free_content_fifo);

      // Build an SSL context
      cdtls->ctx = initialize_ctx(KEYFILE_CLIENT, PASSWORD_CLIENT, DTLSv1_client_method());
      SSL_CTX_set_session_id_context(cdtls->ctx, &(cdtls->session_id), sizeof(cdtls->session_id));
      SSL_CTX_set_mode(cdtls->ctx, SSL_MODE_AUTO_RETRY);
      SSL_CTX_set_read_ahead(cdtls->ctx, 1);

      // Build the SSL tunnel
      cdtls->ssl = SSL_new(cdtls->ctx);

      // Build a BIO based on the user socket, and use it for read/write
      cdtls->sbio = BIO_new_dgram(socket, BIO_NOCLOSE);
      BIO_dgram_set_peer(cdtls->sbio, dest_addr);

      SSL_set_bio(cdtls->ssl, cdtls->sbio, cdtls->sbio);

      cdtls->sbio->cb_arg = (char*)cdtls;

      // Set the SSL tunnel in connect state
      SSL_set_connect_state(cdtls->ssl);

      // Set the new peer in the socket dtls session array
      tls_hash_set_content(dtls, id, cdtls);
    }

    // Write data to the SSL session
    int r = SSL_write(cdtls->ssl, buffer, length);

    // Check the action result
    switch (SSL_get_error(cdtls->ssl, r))
    {
      // No error, return the size
      case SSL_ERROR_NONE:
	return r;
    }
  }

  return -1;
}

ssize_t write(int fildes, const void *buf, size_t nbyte)
{
  DFPRINT("\n[[[write (%u)]]]\n", fildes);

  ssl_associate *aso = tls_hash_get_content(&hash_table, fildes);

  if (!aso)
    return my_write(fildes, buf, nbyte);

  return sendto(fildes, buf, nbyte, 0, (struct sockaddr*)aso->addr, aso->addr_len);
}

ssize_t read(int fildes, void *buf, size_t nbyte)
{
  DFPRINT("\n[[[read (%u)]]]\n", fildes);

  ssl_associate *aso = tls_hash_get_content(&hash_table, fildes);

  if (!aso)
    return my_read(fildes, buf, nbyte);

  return recvfrom(fildes, buf, nbyte, 0, (struct sockaddr*)aso->addr, &aso->addr_len);
}

/*
** Interlal Functions Implementation
*/
/*!
** Free an associative (aso) object : called internally by tls_hash
**
** @param table The hash table where is the object
** @param content The object to free
*/
static void free_associate(tls_hash *table, void *content)
{
  //DPRINT("\n{{{free_content}}}\n");

  // If we don't have content, do nothing
  if (!content)
    return;

  // The content should be an ssl_associate object
  ssl_associate *aso = content;

  if (aso->skip)
    return;

  // Check the asotype
  if (aso->type == ssl_type_tcp)
  {
    tls_content *tls = aso->tls;

    // Destroy the TLS context
    if (tls && tls->ctx)
      destroy_ctx(tls->ctx);

    // Destroy the TLS session
    if (tls && tls->ssl)
    {
      SSL_shutdown(tls->ssl);
      SSL_free(tls->ssl);
    }

    // Free the TLS context
    free(tls);
  }
  else if (aso->type == ssl_type_udp)
  {
    tls_hash *dtls = aso->dtls;

    // Clean the hash table of dtls session (tls_hash will call free_content_dtls for each remaining object)
    tls_hash_clean(dtls);

    // Free the the DTLS hash table
    free(dtls);
  }

  // Free the aso structure
  free(aso);
}


/*!
**  Free an dtls_content (a peer) object : called internally by tls_hash
**
** @param table The hash table where is the object
** @param content The object to free
*/
static void free_content_dtls(tls_hash *table, void *content)
{
  // If we don't have content, do nothing
  if (!content)
    return;

  // The content should be a dtls_content
  dtls_content *dtls = content;

  // Destroy the DTLS context
  if (dtls->ctx)
    destroy_ctx(dtls->ctx);

  // Destroy the DTLS session
  if (dtls->ssl)
  {
    SSL_shutdown(dtls->ssl);
    SSL_free(dtls->ssl);
  }

  // Clean the fifo of the peer (tls_fifo will call free_content_fifo for each object)
  tls_fifo_clean(&dtls->fifo);

  free(dtls);
}

/*!
** Free an raw data object : called internally by tls_fifo
**
** @param fifo The fifo queue where is the data
** @param data The data to free
** @param size The size of the data
*/
static void free_content_fifo(tls_fifo *fifo, void *data, unsigned size)
{
  if (data)
    free(data);
}

// Taken from bss_sock.c in openssl sources
static int sock_read(BIO *b, char *out, int outl)
{
  DPRINT("\n{{{SSL BIO Read}}}\n");

  int ret = 0;

  if (out != NULL)
  {
    errno = 0;
    ret = my_recvfrom(b->num, out, outl, 0, NULL, NULL);
    BIO_clear_retry_flags(b);
    if (ret <= 0)
    {
      if (BIO_sock_should_retry(ret))
	BIO_set_retry_read(b);
    }
  }
  return(ret);
}

// Taken from bss_sock.c in openssl sources
static int sock_write(BIO *b, const char *in, int inl)
{
  DPRINT("\n{{{SSL BIO Write}}}\n");

  int ret;

  errno = 0;
  ret = my_sendto(b->num, in, inl, 0, NULL, 0);
  BIO_clear_retry_flags(b);
  if (ret <= 0)
  {
    if (BIO_sock_should_retry(ret))
      BIO_set_retry_write(b);
  }
  return ret;
}

// Taken from bss_sock.c in openssl sources
static int sock_puts(BIO *bp, const char *str)
{
  DPRINT("\n{{{SSL BIO Puts}}}\n");
  int n, ret;

  n = strlen(str);
  ret = sock_write(bp, str, n);
  return ret;
}

// Taken from bss_sock.c in openssl sources
static int BIO_dgram_should_retry(int i)
{
  int err;

  if ((i == 0) || (i == -1))
  {
    err = errno;

    return(BIO_dgram_non_fatal_error(err));
  }
  return 0;
}

/*!
** Ask libc to receive a datagram on a socket. Use the remote address given by the system to retrieve the dtls_content (DTLS session) of the remote peer.
**
** If the peer doesn't exist, it is created and the DTLS session information are initialized. The peer is added on the peer array of the socket (via the aso).
**
** If the peer exist, simply return it.
**
** The frame read to know the address is backed up on the fifo of the peer (with a maximum of 1 Mo of data, else old datagrams are dropped).
**
** The address read is backed up on the peer to be given-back to the user after.
**
** At this time, we can't know the size of the next datagram, so we give a big value and except that the next datagram will be smaller than this.
**
** @param socket The socket on wich we should read.
**
** @return The dtls_content (peer) from where the datagram was send, and NULL if we have an error
*/
static dtls_content *dtls_read(int socket)
{
  DPRINT("\n{{{SSL DTLS Read}}}\n");

  char			buffer[50000];
  struct sockaddr_in	remote;
  socklen_t		remote_len = sizeof(remote);
  ssl_associate		*aso;
  int			ret;

  // Get the associate (aso) structure with this socket
  DPRINT("-> get the aso structure\n");
  aso = tls_hash_get_content(&hash_table, socket);
  if (!aso)
    return NULL;

  // Check the aso type
  DPRINT("-> check the aso type\n");
  if (aso->type != ssl_type_udp || !aso->dtls)
    return NULL;

  // Receive a datagrame (use a big buffer)
  DPRINT("-> try to receive data\n");
  ret = my_recvfrom(socket, &buffer, sizeof(buffer), 0, (struct sockaddr *)&remote, &remote_len);
  if (ret <= 0)
    return NULL;

  // Check that we have really a sockaddr_in
  DPRINT("-> check the remote address structure\n");
  if (remote_len != sizeof(remote) || remote.sin_family != AF_INET)
    return NULL;

  // Get the dtls_content associated with this peer
  long long	id = peer_key(&remote);
  dtls_content	*cdtls = tls_hash_get_content(aso->dtls, id);

  // If the dtls_content doesn't exist, it's the first time that we communicate with this peer
  if (!cdtls)
  {
    DPRINT("*** build a new peer\n");
    cdtls = calloc(1, sizeof(dtls_content));

    // Initialize the FIFO
    tls_fifo_init(&cdtls->fifo, free_content_fifo);

    // Initialize SSL context
    cdtls->ctx = initialize_ctx(KEYFILE_CLIENT, PASSWORD_CLIENT, DTLSv1_server_method());
    SSL_CTX_set_session_id_context(cdtls->ctx, &(cdtls->session_id), sizeof(cdtls->session_id));
    SSL_CTX_set_mode(cdtls->ctx, SSL_MODE_AUTO_RETRY);
    SSL_CTX_set_read_ahead(cdtls->ctx, 1);

    // Build the SSL tunnel
    cdtls->ssl = SSL_new(cdtls->ctx);

    // Build a BIO based on the user socket, and use it for read/write
    cdtls->sbio = BIO_new_dgram(socket, BIO_NOCLOSE);
    SSL_set_bio(cdtls->ssl, cdtls->sbio, cdtls->sbio);

    // Do something dirty : use an mostly unused var in the BIO structure
    // to hold our cdtls structure
    cdtls->sbio->cb_arg = (char*)cdtls;

    // Set the SSL tunnel in accept state
    SSL_set_accept_state(cdtls->ssl);

    // Store the new peer
    tls_hash_set_content(aso->dtls, id, cdtls);
  }

  // Push the datagrame in the dtls data fifo
  void *data = malloc(ret);
  memcpy(data, buffer, ret);
  tls_fifo_push(&cdtls->fifo, data, ret);

  // If we are more than 1 Mo of data, drop data until we are less than that
  while (tls_fifo_size(&cdtls->fifo) > 1024 * 1024 &&
	 tls_fifo_pop(&cdtls->fifo, &data, NULL))
  {
    DPRINT("-> too much data in FIFO : deleting a frame\n");
    free(data);
  }

  // Copy the address
  cdtls->remote_len = remote_len;
  memcpy((void*)&cdtls->remote, (void*)&remote, remote_len);

  DPRINT("-> normal exit\n");
  return cdtls;
}

// Taken from bss_dgram.c in openssl sources
static int dgram_read(BIO *b, char *out, int outl)
{
  DFPRINT("\n{{{SSL DBIO Read (%i)}}}\n", outl);

  int ret = 0;
  bio_dgram_data *data = (bio_dgram_data *)b->ptr;

  // Something dirty : we use a mostly unused var in the BIO structure
  // to retrieve our peer session object for this BIO (and so for the DTLS session).
  dtls_content	*cdtls = (dtls_content *)b->cb_arg;

  struct sockaddr peer;
  int peerlen = sizeof(peer);

  if (out != NULL && cdtls)
  {
    errno = 0;
    memset(&peer, 0x00, peerlen);

    // Loop until we get a datagram
    while (1)
    {
      void	*data;
      unsigned	data_size;

      // Try pop a previoulsy pushed datagram on the dtls_content (peer)
      DPRINT("-> get the buffer\n");
      if (tls_fifo_pop(&cdtls->fifo, &data, &data_size))
      {
	// If we have a datagram, use it
	DFPRINT("-> there is a buffer of %u\n", data_size);
	ret = MIN(data_size, outl);

	DFPRINT("-> copy data %u\n", ret);
	memcpy(out, data, ret);
	free(data);
	break;
      }
      else
      {
	// If we don't have remaining datagram in the dtls_content
	dtls_content *tmp = NULL;

	// Try to get a datagram, and retry until we get a datagram sended by the peer of this BIO
	// (the other datagrams are backed-up on the related peer)
	while (tmp != cdtls)
	{
	  DPRINT("-> try to get a new datagram");
	  tmp = dtls_read(b->num);
	  if (!tmp)
	  {
	    DPRINT("-> failed to get a new datagram");
	    break;
	  }
	}
      }
    }

    // Copy the datagram to the caller
    memcpy((void*)&peer, (void*)&cdtls->remote, cdtls->remote_len);
    peerlen = cdtls->remote_len;

#if DEBUG
    struct sockaddr_in *deb = (struct sockaddr_in *)&peer;
    DFPRINT("-> received from %s:%u\n", inet_ntoa(deb->sin_addr), ntohs(deb->sin_port));
#endif

    if (!data->connected && ret > 0)
      BIO_ctrl(b, BIO_CTRL_DGRAM_CONNECT, 0, &peer);

    BIO_clear_retry_flags(b);
    if (ret <= 0)
    {
      DPRINT("-> error\n");
      if (BIO_dgram_should_retry(ret))
      {
	BIO_set_retry_read(b);
	data->_errno = errno;
      }
    }
  }

  DPRINT("-> normal exit\n");
  return ret;
}

// Taken from bss_dgram.c in openssl sources
static int dgram_write(BIO *b, const char *in, int inl)
{
  DFPRINT("\n{{{SSL DBIO Write (%i)}}}\n", inl);

  int ret;
  bio_dgram_data *data = (bio_dgram_data *)b->ptr;

#if DEBUG
  struct sockaddr_in *deb = (struct sockaddr_in *)&data->peer;
  DFPRINT("-> send to %s:%u\n", inet_ntoa(deb->sin_addr), ntohs(deb->sin_port));
#endif

  errno = 0;

   if (data->connected)
    ret = my_sendto(b->num, in, inl, 0, NULL, 0);
  else
    ret = my_sendto(b->num, in, inl, 0, &data->peer, sizeof(data->peer));

  BIO_clear_retry_flags(b);
  if (ret <= 0)
  {
    if (BIO_sock_should_retry(ret))
    {
      BIO_set_retry_write(b);
      data->_errno = errno;
    }
  }
  return ret;
}

// Taken from bss_dgram.c in openssl sources
static int dgram_puts(BIO *bp, const char *str)
{
  DPRINT("\n{{{SSL DBIO Puts}}}\n");

  int n, ret;

  n = strlen(str);
  ret = dgram_write(bp, str, n);

  return ret;
}

/*!
** Build a key from a IPv4 address object (this lib doesn't work on IPv6 !) : IP + port
**
** @param peer_key The address object
**
** @return A key to use in a tls_hash table
*/
static inline long long (peer_key)(struct sockaddr_in *address)
{
  if (!address)
    return 0;

  long long key = 0;
  in_port_t port = address->sin_port;
  in_addr_t ip = address->sin_addr.s_addr;

  key = key | (ip << (sizeof(in_port_t) * 8));
  key = key | port;

  return key;
}
